/*
 * Copyright (c) 2016-2019, National Institute of Informatics
 */
/*
 * Licensed to the University Corporation for Advanced Internet Development, 
 * Inc. (UCAID) under one or more contributor license agreements.  See the 
 * NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The UCAID licenses this file to You under the Apache 
 * License, Version 2.0 (the "License"); you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute.filter.impl;

import java.util.Collection;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import jp.gakunin.idp.attribute.UApproveJPIdPAttribute;
import net.shibboleth.idp.attribute.filter.AttributeFilter;
import net.shibboleth.idp.attribute.filter.AttributeFilterException;
import net.shibboleth.idp.attribute.filter.AttributeFilterPolicy;
import net.shibboleth.idp.attribute.filter.AttributeRule;
import net.shibboleth.idp.attribute.filter.Matcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.AbstractComposedMatcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.AndMatcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.OrMatcher;
import jp.gakunin.idp.attribute.filter.matcher.saml.impl.AttributeInMetadataMatcher;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterContext;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterWorkContext;
import jp.gakunin.idp.attribute.filter.context.AttributeFilterOptionalContext;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NullableElements;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.logic.Constraint;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Service that filters out attributes and values based upon loaded policies. */
@ThreadSafe
public class AttributeFilterImpl extends net.shibboleth.idp.attribute.filter.impl.AttributeFilterImpl implements AttributeFilter {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeFilterImpl.class);

    /**
     * {@inheritDoc}
     */
    public AttributeFilterImpl(@Nonnull @NotEmpty String engineId,
            @Nullable @NullableElements final Collection<AttributeFilterPolicy> policies) {
        super(engineId, policies);
    }

    /**
     * {@inheritDoc}
     */
    @Override public void filterAttributes(@Nonnull final AttributeFilterContext filterContext)
            throws AttributeFilterException {
        ComponentSupport.ifNotInitializedThrowUninitializedComponentException(this);
        ComponentSupport.ifDestroyedThrowDestroyedComponentException(this);

        Constraint.isNotNull(filterContext, "Attribute filter context can not be null");
        Map<String, IdPAttribute> prefilteredAttributes = filterContext.getPrefilteredIdPAttributes();

        // Create work context to hold intermediate results.
        final AttributeFilterWorkContext filterWorkContextIncludedOption =
            filterContext.getSubcontext(AttributeFilterWorkContext.class, true);

        // Create optional context to hold intermediate results.
        final AttributeFilterOptionalContext filterOptionalContext = 
            filterContext.getSubcontext(AttributeFilterOptionalContext.class, true);

        log.debug("Beginning process of filtering the following {} attributes: {}", new Object[] {
                prefilteredAttributes.size(), prefilteredAttributes.keySet(),});

        // Apply policies with optional values.
        final List<AttributeFilterPolicy> policies = getFilterPolicies();
        for (AttributeFilterPolicy policy : policies) {
            policy.apply(filterContext);
        }

        // Store results at first policy application.
        final Map<String, Set<IdPAttributeValue>> permittedAttributeValueMap =
            filterWorkContextIncludedOption.getPermittedIdPAttributeValues();
        final Map<String, Set<IdPAttributeValue>> deniedAttributeValueMap =
            filterWorkContextIncludedOption.getDeniedAttributeValues();

        // Remove work context at first policy application and create the context again.
        filterContext.removeSubcontext(AttributeFilterWorkContext.class);
        final AttributeFilterWorkContext filterWorkContextExcludedOption =
            filterContext.getSubcontext(AttributeFilterWorkContext.class, true);

        // Set mandatory value mode is on.
        filterOptionalContext.setMandatoryValueMode(true);

        // Apply policies without optional values.
        for (AttributeFilterPolicy policy : policies) {
            policy.apply(filterContext);
        }

        UApproveJPIdPAttribute filteredAttribute;
        for (String attributeId : filterContext.getPrefilteredIdPAttributes().keySet()) {
            final Collection permittedAttributeValues = permittedAttributeValueMap.get(attributeId);
            if (null == permittedAttributeValues || permittedAttributeValues.isEmpty()) {
                log.debug("Attribute filtering engine '{}': no policy permitted release of attribute {} values", getId(),
                          attributeId);
                continue;
            }

            final Collection mandatoryPermittedAttributeValues = 
                filterWorkContextExcludedOption.getPermittedIdPAttributeValues().get(attributeId);

            // Remove denied values from permitted values.
            final Collection deniedAttributeValues = deniedAttributeValueMap.get(attributeId);
            if (null != deniedAttributeValues && !deniedAttributeValues.isEmpty()) {
                permittedAttributeValues.removeAll(deniedAttributeValues);

                if (null != mandatoryPermittedAttributeValues) {
                    mandatoryPermittedAttributeValues.removeAll(deniedAttributeValues);
                }

                if (permittedAttributeValues.isEmpty()) {
                    log.debug("Attribute filtering engine '{}': deny policies filtered out all values for attribute '{}'",
                              getId(), attributeId);
                    continue;
                }
            }
            log.debug("Attribute filtering engine '{}': {} values for attribute '{}' remained after filtering",
                      new Object[] {getId(), permittedAttributeValues.size(), attributeId,});

            /* 
             * When one or more mandatory values exists, mark this attribute as mandatory attribute.
             * Otherwise mark it as optional attribute.
             */
            boolean optional = false;
            if (null == mandatoryPermittedAttributeValues || mandatoryPermittedAttributeValues.isEmpty()) {
                optional = true;
            }
            log.debug("Attribute filtering engine '{}': attribute {} is {}", getId(), attributeId,
		      optional ? "optional" : "mandatory");

            try {
                filteredAttribute = new UApproveJPIdPAttribute(attributeId);
                filteredAttribute.setIdPAttributeProperties(prefilteredAttributes.get(attributeId).clone());
            } catch (CloneNotSupportedException e) {
                throw new AttributeFilterException(e);
            }
            filteredAttribute.setValues(permittedAttributeValues);
            filteredAttribute.setOptional(optional);
            filterContext.getFilteredIdPAttributes().put(filteredAttribute.getId(), filteredAttribute);
        }
    }
}
