/*
 * Copyright (c) 2016, National Institute of Informatics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.consent.logic.impl;

import java.util.Enumeration;
import java.util.Locale;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.Collections;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import net.shibboleth.idp.ui.context.RelyingPartyUIContext;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.AttributeEncoder;
import net.shibboleth.idp.saml.profile.context.navigate.SAMLMetadataContextLookupFunction;
import net.shibboleth.idp.saml.attribute.encoding.SAML1AttributeEncoder;
import net.shibboleth.idp.saml.attribute.encoding.SAML2AttributeEncoder;

import jp.gakunin.idp.saml.xmlobject.UApproveJPMetadata;
import jp.gakunin.idp.saml.xmlobject.RequestedAttributeExtension;
import jp.gakunin.idp.saml.xmlobject.Description;

import org.opensaml.core.xml.XMLObject;
import org.opensaml.saml.common.messaging.context.SAMLMetadataContext;
import org.opensaml.saml.common.messaging.context.AttributeConsumingServiceContext;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml.saml2.metadata.RequestedAttribute;
import org.opensaml.saml.saml2.metadata.SPSSODescriptor;
import org.opensaml.saml.saml2.metadata.RoleDescriptor;
import org.opensaml.saml.saml2.metadata.Extensions;
import org.opensaml.profile.context.ProfileRequestContext;

import com.google.common.collect.ImmutableMap;
import com.google.common.base.Function;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Function which returns the locale-aware intended use of an attribute,
 */  
public class AttributeIntendedUseFunction implements Function<IdPAttribute, String> {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeIntendedUseFunction.class);

    /** The languages that this the Operator want to fall back to. */
    @Nonnull private List<String> defaultLanguages;

    /** The locales that this browser wants to know about.  */
    @Nonnull private List<Locale> browserLocales;

    /** The locales that this the bean needs to look at. */
    @Nonnull private List<Locale> usableLocales;

    /** RequestAttributes in Metadata. */
    private List<RequestedAttribute> requested;

    /** Map a friendlyName to IntendedUses. */
    private Map<String, Map<Locale, String>> friendlyNameToIntendedUses;

    /**
      * Constructor.
      */
    public AttributeIntendedUseFunction() {
        defaultLanguages = new ArrayList<>();
        browserLocales = new ArrayList<>();
        usableLocales = new ArrayList<>();
        requested = new ArrayList<>();
        friendlyNameToIntendedUses = new HashMap<>();
    }

    /**
     * Initializer.
     * Find descriptions of attribute from the SP's metadata.
     * Set the browser locales.
     * 
     * @param profileContext the current profile request context
     * @param locales the locales that this browser wants to know about
     * 
     * @return this function, or null
     */
    @Nullable public final AttributeIntendedUseFunction initialize(@Nullable final ProfileRequestContext profileContext, final Enumeration<Locale> locales) {
        if (profileContext == null || locales == null) {
            return null;
        }

        SAMLMetadataContextLookupFunction samlMetadataLookupStrategy = new SAMLMetadataContextLookupFunction();
        SAMLMetadataContext metadataContext = samlMetadataLookupStrategy.apply(profileContext);
        if (metadataContext == null) {
            log.debug("SAML Metadata context not found.");
            return null;
        }

        requested = getRequestedAttributes(metadataContext);
        friendlyNameToIntendedUses = getFriendlyNameToIntendedUsesMap(metadataContext);

        setBrowserLocales(locales);
        return this;
    }

    /**
     * Get requested attributes in metadata.
     *
     * @param metadataContext saml metadata context
     *
     * @return list of requested attributes
     */
    private final List<RequestedAttribute> getRequestedAttributes(final SAMLMetadataContext metadataContext) {
        final AttributeConsumingServiceContext acsCtx =
                metadataContext.getSubcontext(AttributeConsumingServiceContext.class);
        if (acsCtx == null) {
            log.debug("Attribute consuming service context not found.");
            return Collections.emptyList();
        }

        final AttributeConsumingService service = acsCtx.getAttributeConsumingService();
        if (service == null) {
            log.debug("Attribute consuming service not found.");
            return Collections.emptyList();
        }

        return service.getRequestAttributes();
    }

    /**
     * Map to IntendedUses from FriendlyName.
     *
     * @param metadataContext saml metadata context 
     *
     * @return generated map
     */
    private final Map<String, Map<Locale, String>> getFriendlyNameToIntendedUsesMap(final SAMLMetadataContext metadataContext) {
        SPSSODescriptor spSSODescriptor = null;
        final RoleDescriptor roleDescriptor = metadataContext.getRoleDescriptor();
        if (roleDescriptor instanceof SPSSODescriptor) {
            spSSODescriptor = (SPSSODescriptor) roleDescriptor;
        }
        if (spSSODescriptor == null) {
            log.debug("SPSSODescriptor not found.");
            return Collections.emptyMap();
        }

        final Extensions exts = spSSODescriptor.getExtensions();
        if (exts == null) {
            log.debug("Extensions not found.");
            return Collections.emptyMap();
        }

        Map<String, Map<Locale, String>> intendedUsesMap = new HashMap<>();
        for (XMLObject object : exts.getOrderedChildren()) {
            if (!(object instanceof RequestedAttributeExtension)) {
                log.trace("Not RequestedAttributeExtension");
                continue;
            }
            RequestedAttributeExtension raExtention = (RequestedAttributeExtension) object;
            String name = raExtention.getFriendlyName();
            if (name == null || name.isEmpty()) {
                log.trace("FriendlyName is null or empty");
                continue;
            }
            List<Description> raDescs = raExtention.getDescriptions();
            if (raDescs == null || raDescs.isEmpty()) {
                log.trace("Desctiptions is null or empty");
                continue;
            }
            Map<Locale, String> intendedUses = new HashMap<>();
            for (final Description desc : raDescs) {
                if (desc.getXMLLang() == null) {
                    log.trace("XMLLang is null");
                    continue;
                }
                intendedUses.put(Locale.forLanguageTag(desc.getXMLLang()), desc.getValue());
            }
            if (!intendedUses.isEmpty()) {
                log.debug("FriendlyName {} found Descriptions in RequestedAttributeExtension: {}", name, intendedUses);
                intendedUsesMap.put(name, intendedUses);
            }
        }
        return ImmutableMap.copyOf(intendedUsesMap);
    }


    /**
     * Set the browser locales.
     *
     * @param locales the locales to set
     */
    @Nullable private void setBrowserLocales(final Enumeration<Locale> locales) {
        if (locales == null) {
            log.trace("locales is null");
            return;
        }
        final List<Locale> newLocales = new ArrayList<>();
        while (locales.hasMoreElements()) {
            newLocales.add(locales.nextElement());
        }
        browserLocales = newLocales;
        makeUsableLocales();
    }

    /**
     * Set the default languages.
     *
     * @param langs the languages to set
     */
    @Nullable public void setDefaultLanguages(final List<String> langs) {
        if (langs == null) {
            log.trace("languages is null");
            return;
        }
        defaultLanguages = langs;
        makeUsableLocales();
    }

    /**
     *  Construct the usableLocales from the {@link #browserLocales} and the {@link #defaultLanguages}.
     */
    private void makeUsableLocales() {
        List<Locale> newLocales = new ArrayList<>();

        newLocales = browserLocales;
        for (Locale locale : browserLocales) {
            log.trace("set browser locales: {}", locale);
        }

        for (final String s : defaultLanguages) {
            if (s != null) {
                newLocales.add(new Locale(s));
                log.trace("set default locales: {}", new Locale(s));
            }
        }
        usableLocales = newLocales;
    }

    /** {@inheritDoc} */
    @Nullable public String apply(@Nullable final IdPAttribute input) {
        if (input == null) {
            log.trace("Attribute is null");
            return null;
        }

        RequestedAttribute requestedAttribute = findInMetadata(input);
        if (requestedAttribute == null) {
            log.debug("Attribute {} not found in metadata", input.getId());
            return null;
        }

        String toBeDisplayed = getIntendedUseFromExtensions(requestedAttribute);
        if (toBeDisplayed != null) {
            log.debug("Attribute {} found Description in Extensions: {}", input.getId(), toBeDisplayed);
            return toBeDisplayed;
        } else {
            log.debug("Attribute {} not found Description in Extensions", input.getId(), toBeDisplayed);
        }

        toBeDisplayed = getIntendedUseFromRequestedAttribute(requestedAttribute);
        if (toBeDisplayed != null) {
            log.debug("Attribute {} found Description in RequestedAttribute: {}", input.getId(), toBeDisplayed);
        } else {
            log.debug("Attribute {} not found Description", input.getId());
        }
        return toBeDisplayed;
    }

    /**
     * Get the intended use from Extensions in metadata.
     *
     * @param requestedAttribute requested attribute
     *
     * @return a intended use of attribute, or null
     */
    @Nullable private final String getIntendedUseFromExtensions(@Nonnull final RequestedAttribute requestedAttribute) {
        final String name = requestedAttribute.getFriendlyName();
        if (!friendlyNameToIntendedUses.containsKey(name)) {
            log.debug("FriendlyName {} not found Desctiption in Extensions", name);
            return null;
        }

        final Map<Locale, String> intendedUses = friendlyNameToIntendedUses.get(name);
        if (intendedUses != null) {
            for (Locale locale : usableLocales) {
                if (intendedUses.containsKey(locale)) {
                    return intendedUses.get(locale);
                }
                if (intendedUses.containsKey(Locale.forLanguageTag(locale.getLanguage()))) {
                    return intendedUses.get(Locale.forLanguageTag(locale.getLanguage()));
                }
            }
        }
        return null;
    }

    /**
     * Get the intended use from RequestedAttribute in metadata.
     *
     * @param requestedAttribute requested attribute
     *
     * @return a intended use of attribute, or null
     */
    @Nullable private final String getIntendedUseFromRequestedAttribute(@Nonnull final RequestedAttribute requestedAttribute) {
        return requestedAttribute.getUnknownAttributes().get(UApproveJPMetadata.ATTRDESC_NAME);
    }

    /**
     * Locates a {@link RequestedAttribute} object in metadata that matches the eventual "to be encoded" name
     * represented by the encoder.
     *
     * @param encoder the attribute encoder to base the search on
     *
     * @return a matching RequestedAttribute, or null
     */
    @Nullable private RequestedAttribute findInMetadata(@Nonnull final IdPAttribute attribute) {
        final Set<AttributeEncoder<?>> encoders = attribute.getEncoders();
        for (final AttributeEncoder<?> encoder : encoders) {
            String name = null;
            String qualifier = null;
            if (encoder instanceof SAML2AttributeEncoder) {
                name = ((SAML2AttributeEncoder) encoder).getName();
                qualifier = ((SAML2AttributeEncoder) encoder).getNameFormat();
            } else if (encoder instanceof SAML1AttributeEncoder) {
                name = ((SAML1AttributeEncoder) encoder).getName();
                qualifier = ((SAML1AttributeEncoder) encoder).getNamespace();
            }
    
            for (final RequestedAttribute attr : requested) {
                if (attr.getName().equals(name)) {
                    final String format = attr.getNameFormat();
                    if (qualifier == null || format == null || format.equals(Attribute.UNSPECIFIED)
                            || qualifier.equals(format)) {
                        return attr;
                    }
                }
            }
        }
        return null;
    }

}
